package edu.gmu.trimdroid.soot.callgraph;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import edu.gmu.trimdroid.soot.callgraph.CGEdge.EdgeType;
import edu.gmu.trimdroid.soot.utils.Const;
import edu.gmu.trimdroid.soot.utils.Util;
import soot.PatchingChain;
import soot.Scene;
import soot.SootClass;
import soot.SootMethod;
import soot.SootMethodRef;
import soot.Unit;
import soot.Value;
import soot.ValueBox;
import soot.jimple.InvokeExpr;

public class IPCNodeExtractor implements NodeExtractor {

	@Override
	public List<CGNode> extractNode(Unit unit, PatchingChain<Unit> units, SootMethod thisMethod, CGNode fromNode, SootClass rClazz) {
		List<SootMethod> intentTargets = findExplicitIntentTarget(unit, thisMethod, units);
		if (intentTargets == null)
			return null;
		List<CGNode> result = new ArrayList<>();
		for (SootMethod intentTarget : intentTargets) {
			result.add(connectNodes(intentTarget, fromNode));
		}
		return result;
	}

	private static CGNode connectNodes(SootMethod intentTarget, CGNode fromNode) {
		CGNode toNode = new CGNode(intentTarget);
		CGEdge edge = new CGEdge(EdgeType.IPC);
		CGNode.connect(fromNode, toNode, edge, true);
		return toNode;
	}

	public static List<CGNode> extractIPCNode(Unit unit, SootMethod thisMethod, PatchingChain<Unit> units) {
		List<CGNode> nodes = new ArrayList<>();
		List<SootMethod> intentTarget = findExplicitIntentTarget(unit, thisMethod, units);
		if (intentTarget == null)
			return nodes;
		for (SootMethod method : intentTarget) {
			nodes.add(new CGNode(method));
		}
		return nodes;
	}

	private static List<SootMethod> findExplicitIntentTarget(Unit unit, SootMethod thisMethod, PatchingChain<Unit> units) {
		List<ValueBox> useBoxes = unit.getUseBoxes();
		for (ValueBox valueBox : useBoxes) {
			Value value = valueBox.getValue();
			if (value instanceof InvokeExpr) {
				InvokeExpr invokeExpr = (InvokeExpr) value;
				SootMethodRef methodRef = invokeExpr.getMethodRef();
				String methodName = methodRef.name();
				if (invokeExpr.getMethod().getDeclaringClass().getType().getClassName().contains("Intent")) {
					String className = null;
					if (methodName.equals(Const.METHOD_INIT)) {
						int argCount = invokeExpr.getArgCount();
						if (argCount == 4 || (argCount == 2 && !invokeExpr.getArg(0).getType().toString().equals(Const.TYPE_STRING))) {
							Value lastArg = invokeExpr.getArg(argCount - 1);
							className = sootClassNameToJavaClassName(lastArg.toString());
						}
					} else if (Arrays.asList(Const.TARGET_SETTER_METHOD_NAMES).contains(methodName)) {
						Value arg2 = invokeExpr.getArg(1);
						className = extractClassNameFromArguments(arg2, thisMethod);
					} else if (methodName.equals(Const.METHOD_SET_COMPONENT)) {
						Unit compDef = Util.findClosestUsePointBackward(unit, invokeExpr.getArg(0), units);
						InvokeExpr compMethod = Util.getInvokedMethod(compDef);
						if (compMethod != null && compMethod.getMethod().getName().equals(Const.METHOD_INIT) && compMethod.getArgCount() == 2) {
							className = extractClassNameFromArguments(compMethod.getArg(1), thisMethod);
						}
					}
					if (className == null)
						return null;
					List<SootMethod> entryPointMethods = new ArrayList<>();
					SootClass intentTarget = Scene.v().getSootClass(className);
					for (String entryPoint : Const.ENTRY_POINT_METHODS) {
						try {
							SootMethod entryPointMethod = intentTarget.getMethodByName(entryPoint);
							entryPointMethods.add(entryPointMethod);
						} catch (Exception e) {

						}
					}
					return entryPointMethods;
				}
			}
		}
		return null;
	}

	private static String sootClassNameToJavaClassName(String className) {
		if (className == null)
			return null;
		return className.substring(7, className.length() - 1).replaceAll("/", ".");
	}

	private static String extractClassNameFromArguments(Value arg, SootMethod method) {
		if (arg.getType().toString().contains("Class"))
			return sootClassNameToJavaClassName(arg.toString());
		return method.getDeclaringClass().getPackageName() + "." + arg.toString();
	}
}
